/*
 * Copyright 2006 Takahiro Nakamura.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package woolpack.utils;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Builder パターンを適用した{@link LinkedHashMap}です。
 * <br/>適用しているデザインパターン：Builder。
 * @author nakamura
 *
 * @param <K>
 * @param <V>
 */
public class BuildableLinkedHashMap<K, V> extends LinkedHashMap<K, V> {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	private transient V beforeValue;
	
	public BuildableLinkedHashMap() {
		super();
	}

	/**
	 * {@link #put(Object, Object)}を実行しこのオブジェクトへの参照を返します。
	 * @param key 追加するキー。
	 * @param value 追加する値。
	 * @return オブジェクトへの参照。
	 */
	public BuildableLinkedHashMap<K, V> map(final K key, final V value) {
		put(key, value);
		beforeValue = value;
		return this;
	}

	/**
	 * {@link #put(Object, Object)}を実行しこのオブジェクトへの参照を返します。
	 * 前回実行した{@link #map(Object, Object)}の値を値として格納します。
	 * @param key 追加するキー。
	 * @return オブジェクトへの参照。
	 */
	public BuildableLinkedHashMap<K, V> map(final K key) {
		put(key, beforeValue);
		return this;
	}
	
	/**
	 * {@link #putAll(Map)}を実行しこのオブジェクトへの参照を返します。
	 * @param t 追加するマップ。
	 * @return オブジェクトへの参照。
	 */
	public BuildableLinkedHashMap<K, V> merge(final Map<? extends K, ? extends V> t) {
		putAll(t);
		return this;
	}
}
